import kotlinx.cli.*
import java.io.File

// Pre setup some regexes.
val methodRegex = Regex("""(?<accessibility>public|private|internal|protected)\s(async\s)?((static|virtual|abstract|override|new)\s)?(async\s)?(?<returnType>[a-zA-Z0-9<>]*)\s(?<methodName>[A-Za-z_0-9<>, ]*)\((?<arguments>[A-Za-z_0-9\[\]<>, ]*)\)""")
val methodArgumentRegex = Regex("""((this|out|ref|in) )?(?<argument>([a-zA-Z0-9]+(\[,*])? [a-zA-Z0-9]+)|([a-zA-Z0-9]+<([a-zA-Z0-9<>]+(\[,*])?(, )?)*>(\[,*])? ?[a-zA-Z0-9]+))""")
val methodCommentArgumentRegex = Regex("""<param name="(?<name>[a-zA-Z0-9]*)">(?<description>.*)</param>""")

/**
 * Entry point of the application.
 * @param args The command line arguments.
 */
fun main(args: Array<String>) {
    // Parse command line arguments.
    val parser = ArgParser("CSharpCommentChecker")
    val folders by parser.argument(ArgType.String, description = "Folders to check").vararg()
    parser.parse(args)

    // Find all files that should be checked.
    val files = folders
        .flatMap { File(it).walkTopDown() }
        .filter { it.isFile && it.extension == "cs" }

    // Check every file and collect their issues.
    val issues = files
        .flatMap(::checkFile)

    // Print all issues to stdout.
    issues.map(Issue::toPrintable).forEach(::println)

    // If there were issues exit with a non-zero status code.
    if (issues.isNotEmpty())
        kotlin.system.exitProcess(1)
}


/**
 * Check the given file for issues and return a these issues if any in a list.
 * @param fileToCheck The file to check for issues.
 * @return A list of [Issue] objects each representing exactly one issue that exists in the file.
 */
fun checkFile(fileToCheck: File): List<Issue> {
    // Check both method and inline comments on the file and combine their issues.
    return checkMethodComments(fileToCheck) +
            checkInlineComments(fileToCheck)
}

/**
 * Check the given file for issues with the inline comments.
 * @param fileToCheck The file to check for issues.
 * @return A list of [Issue] objects each representing exactly one issue that exists in the file with regard to
 * inline comments.
 */
fun checkInlineComments(fileToCheck: File): List<Issue> {
    // Read all lines in the file and trim any extra whitespace from them.
    val fileLines = fileToCheck.readLines()
        .map { it.trim() }

    // Find all lines that include a comment and are not a method comment.
    val commentLines = fileLines
        .mapIndexedNotNull { i, content -> if (content.startsWith("//") && !content.startsWith("///")) i else null }

    // Group the found comments into blocks and process each block into its own string.
    val commentBlocks = groupCommentLines(commentLines)
        .map {section -> section
            .map { it to fileLines[it] }
            .map { it.first to it.second.removePrefix("//") }
            .map { it.first to it.second.trim() }
            .reduce { acc, it -> acc.first to "${acc.second}\n${it.second}" }
        }

    // Check for issues in each block and return the list of issues that are found.
    return commentBlocks
        .filterNot { "<exempt>" in it.second }
        .filter { !it.second.first().isUpperCase() || !it.second.endsWith(".") }
        .map { Issue.inlineComment(fileToCheck, it.first) }
}

/**
 * Find groups of integers that follow each other without gaps to create comment blocks.
 * @param commentLines The list of integers to group.
 * @return A list of lists containing integers. Each list represents a grouping.
 */
fun groupCommentLines(commentLines: List<Int>): List<List<Int>> {
    var currentLine = commentLines.firstOrNull() ?: return emptyList()

    val groups = mutableListOf<List<Int>>()
    var currentGroup = mutableListOf(currentLine)

    // Loop through every integer in the given list except for the first since that is already processed above.
    commentLines.drop(1).forEach { line ->
        // If the lasted checked one is one lower than this one
        if (currentLine + 1 == line)
            // Add this line to the current group.
            currentGroup += line
        else {
            // Otherwise, store the current group and start a new one with this as the first entry.
            groups += currentGroup
            currentGroup = mutableListOf(line)
        }

        // Finally, update the current line to this one.
        currentLine = line
    }

    // If the current group isn't empty store it with the rest of the groups.
    if (currentGroup.isNotEmpty())
        groups += currentGroup

    // And return the list of all groups.
    return groups
}

/**
 * Check the given file for issues with the method comments.
 * @param fileToCheck The file to check for issues.
 * @return A list of [Issue] objects each representing exactly one issue that exists in the file with regard to
 * method comments.
 */
fun checkMethodComments(fileToCheck: File): List<Issue> {
    // Find all lines that contain methods in the file.
    val methodLines = fileToCheck
        .readLines()
        .map { it.trim() }
        .mapIndexedNotNull { i, content -> if (content.matches(methodRegex)) i else null }

    // Check each of these methods for issues and combine those into a single list to return.
    return methodLines
        .flatMap { checkMethod(fileToCheck, it) }
}

/**
 * Check the given method for issues with regard to its method comment.
 * @param file The file that the method resides in.
 * @param methodLineNumber The line that the method starts at.
 * @return A list of [Issue] objects each representing exactly one issue with the method comment of this method.
 */
fun checkMethod(file: File, methodLineNumber: Int): List<Issue> {
    // Read all lines in the file and trim any extra whitespace from them.
    val lines = file
        .readLines()
        .map(String::trim)

    val methodLine = lines[methodLineNumber]

    // Record constructors are exempt from rules.
    if (methodLine.contains("record"))
        return emptyList()

    // Find the return type, method name, and arguments of the method.
    val match = methodRegex.find(methodLine) ?: return emptyList()
    val (_, _, _, _, _, returnType, methodName, arguments) = match.destructured

    // Find all lines that contain the method comment and join them together with new lines.
    // Start one above the method because that is where comments start.
    var currentLineNumber = methodLineNumber - 1
    val commentLines = mutableListOf<String>()
    while (lines[currentLineNumber].isMethodComment()) {
        commentLines += lines[currentLineNumber].removePrefix("///")
        currentLineNumber--
    }
    val comment = commentLines.reversed().joinToString("\n")

    // If a block is marked as exempt we don't need to check it.
    if ("<exempt>" in comment)
        return emptyList()

    // Check the summary, arguments, and return comments for issues and combine them into a single list if they exist.
    return checkMethodSummary(file, methodLineNumber, methodName, comment) +
            checkMethodArguments(file, methodLineNumber, methodName, comment, arguments) +
            checkMethodReturn(file, methodLineNumber, methodName, comment, returnType)
}

/**
 * Check for issues with the method summary comment.
 * @param file The file that the method resides in.
 * @param methodLine The line number of the method.
 * @param comment The string containing the full method comment.
 * @return A list containing either one or zero [Issue] objects in case there is a problem with the method summary.
 */
fun checkMethodSummary(file: File, methodLine: Int, methodName: String, comment: String): List<Issue> = when {
    "<summary>" !in comment || "</summary>" !in comment -> listOf(Issue.methodComment(file, methodLine, methodName, "does not have a ${CC.CYAN}summary"))
    """<summary>\s+[A-Z]""".toRegex() !in comment -> listOf(Issue.methodComment(file, methodLine, methodName, "has a summary that doesn't start with a space and a capital letter"))
    """\.\s+</summary>""".toRegex() !in comment -> listOf(Issue.methodComment(file, methodLine, methodName, "has a summary that doesn't end with a dot and a space"))
    else -> emptyList()
}

/**
 * Check for issues with the method argument comments.
 * @param file The file that the method resides in.
 * @param methodLine The line number of the method.
 * @param comment The string containing the full method comment.
 * @param arguments The string containing all the method arguments.
 * @return A list containing exactly one [Issue] object for each issue with the method argument comments.
 */
fun checkMethodArguments(file: File, methodLine: Int, methodName: String, comment: String, arguments: String): List<Issue> {
    if (arguments.isEmpty()) return emptyList()

    // Find all individual arguments and their names.
    val argumentNames = methodArgumentRegex.findAll(arguments)
        .mapNotNull { it.groups["argument"]?.value }
        .filter { it.isNotBlank() }
        .map { it.substringAfterLast(' ') }
        .toList()

    // Find all arguments in the comments.
    val commentArguments = methodCommentArgumentRegex
        .findAll(comment)
        .map { it.groups }

    // Split the found comment arguments into lists of names and descriptions.
    val (commentArgumentNames, commentArgumentDescriptions) = commentArguments
        .map { Pair(it["name"]?.value ?: "", it["description"]?.value ?: "") }
        .unzip()

    // Find missing argument comments.
    val missingArgumentIssues = (argumentNames - commentArgumentNames.toSet())
        .map { Issue.missingMethodArgumentComment(file, methodLine, methodName, it)  }

    // Find redundant argument comments.
    val extraArgumentIssues = (commentArgumentNames - argumentNames.toSet())
        .map { Issue.extraMethodArgumentComment(file, methodLine, methodName, it) }

    // Find invalid argument comments.
    val invalidArgumentDescriptionIssues = commentArgumentNames
        .mapIndexed { i, name ->
            if (commentArgumentDescriptions[i].isValidArgumentDescription())
                null
            else
                Issue.invalidMethodArgumentComment(file, methodLine, methodName, name)
        }.filterNotNull()

    // Combine all issues and return them.
    return missingArgumentIssues + extraArgumentIssues + invalidArgumentDescriptionIssues
}

/**
 * Check for issues with the method return comment.
 * @param file The file that the method resides in.
 * @param methodLine The line number of the method.
 * @param comment The string containing the full method comment.
 * @param returnType The return type of the method.
 * @return A list containing either one or zero [Issue] objects in case there is a problem with the method summary.
 */
fun checkMethodReturn(file: File, methodLine: Int, methodName: String, comment: String, returnType: String): List<Issue> {
    if (returnType == "void")
        return emptyList()

    return when {
        "<returns>" !in comment || "</returns>" !in comment -> listOf(Issue.methodComment(file, methodLine, methodName, "does not have a ${CC.CYAN}return comment"))
        """<returns>\s+[A-Z]""".toRegex() !in comment -> listOf(Issue.methodComment(file, methodLine, methodName, "has a return comment that doesn't start with a space and a capital letter"))
        """\.\s+</returns>""".toRegex() !in comment -> listOf(Issue.methodComment(file, methodLine, methodName, "has a return comment that doesn't end with a dot and a space"))
        else -> emptyList()
    }
}

/**
 * Detect whether the string is valid as part of a method comment.
 * @return True if it is, false otherwise.
 */
private fun String.isMethodComment(): Boolean {
    // Allow both method comment lines and annotations to be counted as method comment lines.
    return this.startsWith("///") || this.startsWith("[")
}

/**
 * Check whether the string is a valid argument description.
 * @return True if it is, false otherwise.
 */
private fun String.isValidArgumentDescription(): Boolean {
    // Check if the string starts with a space and a capital letter and ends with a dot and a space.
    return """^\s+[A-Z].*\.\s+$""".toRegex() matches this
}

data class Issue(
    val file: File,
    val line: Int,
    val description: String
) {
    /**
     * Convert the issue into a printable string representation.
     * @return The printable string representation.
     */
    fun toPrintable(): String {
        return "$description${CC.RESET}\n" +
                "File: ${file.parent}${File.separator}${CC.BLUE}${file.name}:${CC.GREEN}${line + 1}${CC.RESET}\n"
    }

    companion object {
        /**
         * Create an [Issue] object for a method comment issue.
         * @param file The file that the issue is in.
         * @param line The line number of the issue.
         * @param method The name of the method that the issue is related to.
         * @param description A description of the issue.
         * @return The issue that was created.
         */
        fun methodComment(file: File, line: Int, method: String, description: String): Issue =
            Issue(file, line, "${CC.YELLOW}Method $method ${CC.RESET}$description")

        /**
         * Create an [Issue] object for a missing method argument comment.
         * @param file The file that the issue is in.
         * @param line The line number of the issue.
         * @param method The name of the method that the issue is related to.
         * @param argument The argument that is missing.
         * @return The issue that was created.
         */
        fun missingMethodArgumentComment(file: File, line: Int, method: String, argument: String): Issue =
            methodComment(file, line, method, "does not have a parameter comment for ${CC.CYAN}parameter $argument")

        /**
         * Create an [Issue] object for a redundant method argument comment.
         * @param file The file that the issue is in.
         * @param line The line number of the issue.
         * @param method The name of the method that the issue is related to.
         * @param argument The argument that is redundant.
         * @return The issue that was created.
         */
        fun extraMethodArgumentComment(file: File, line: Int, method: String, argument: String): Issue =
            methodComment(file, line, method, "has a redundant parameter comment for ${CC.CYAN}parameter $argument")

        /**
         * Create an [Issue] object for an invalid method argument comment.
         * @param file The file that the issue is in.
         * @param line The line number of the issue.
         * @param method The name of the method that the issue is related to.
         * @param argument The argument whose that is invalid.
         * @return The issue that was created.
         */
        fun invalidMethodArgumentComment(file: File, line: Int, method: String, argument: String): Issue =
            methodComment(file, line, method, "has an invalid parameter comment for ${CC.CYAN}parameter $argument")

        /**
         * Create an [Issue] object for an invalid inline comment.
         * @param file The file that the issue is in.
         * @param line The line number of the issue.
         * @return The issue that was created.
         */
        fun inlineComment(file: File, line: Int): Issue =
            Issue(file, line, "${CC.YELLOW}Comment in ${file.name}${CC.RESET} is invalid")
    }
}